﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.IO.MemoryMappedFiles;
using System.IO;
using System.Diagnostics;
using Curves;
using Catmull_Rom_Spline;
using System.Drawing;
using System.Net;
using System.Net.NetworkInformation;

namespace cloudtreeReference
{
    public class CloudtreeReferences
    {
        // ECEF-LLA 변환용 스태틱 멤버변수
        static double MAJA = 6378137.0; // semi major axis of ref ellipsoid
        static double FLAT = 1.0 / 298.2572235630; // flattening coef of ref ellipsoid.
        // These are derived values from the above WGS­84 values
        static double ESQR = (FLAT * (2.0 - FLAT)); // 1st eccentricity squared
        static double OMES = (1.0 - ESQR);  // 1 minus eccentricity squared
        static double EFOR = (ESQR* ESQR);  // Sqr of the 1st eccentricity squared
        static double ASQR = (MAJA * MAJA); // semi major axis squared = nlmaja**

        // 공유메모리(파일매핑)용 변수
        MemoryMappedFile mapFile = null;
        MemoryMappedViewAccessor accessor = null;

        // 베지어 곡선용 변수
        private int POINTS_ON_CURVE = 1000;
        private double[] ptind = null;
        private BezierCurve bc = new BezierCurve();


        #region 수학함수
        // 유클리딘 거리 함수: x-y 2차원 좌표간 거리 반환.
        public double EuclideanDistance(double src_x, double src_y, double dest_x, double dest_y)
        {
            // x-y 2차원 좌표간 거리 반환.
            return Math.Sqrt(Math.Pow(dest_x - src_x, 2) + Math.Pow(dest_y - src_y, 2));
        }

        // 유클리딘 거리 함수: x-y-z 3차원 좌표간 거리 반환.
        public double EuclideanDistance(double src_x, double src_y, double src_z, double dest_x, double dest_y, double dest_z)
        {
            // x-y-z 3차원 좌표간 거리 반환.
            return Math.Sqrt(Math.Pow(dest_x - src_x, 2) + Math.Pow(dest_y - src_y, 2) + Math.Pow(dest_z - src_z, 2));
        }

        public double Gaussian(double mean, double stdev) // mean을 기준, stdev를 표준편차로 하는 가우시안 분포
        {
            Random rand = new Random();
            int RAND_MIN = 1;
            int RAND_MAX = 1000;

            double U1, U2;
            double x;

            U1 = (double)rand.Next(RAND_MIN, RAND_MAX) / RAND_MAX; // 0과 1사이의 유니폼 변수 
            U2 = (double)rand.Next(RAND_MIN, RAND_MAX) / RAND_MAX; // 0과 1사이의 유니폼 변수 

            x = mean + (Math.Sqrt(-2 * Math.Log(U1)) * Math.Cos(2 * Math.PI * U2)) * stdev;
            return x;
        }
        #endregion


        #region 좌표계
        public double bearingP1toP2(double P1_lat, double P1_lon, double P2_lat, double P2_lon)
        {
            // 현재 위치 : 위도나 경도는 지구 중심을 기반으로 하는 각도이기 때문에 라디안 각도로 변환한다.
            double Cur_Lat_radian = P1_lat * (Math.PI / 180);
            double Cur_Lon_radian = P1_lon * (Math.PI / 180);


            // 목표 위치 : 위도나 경도는 지구 중심을 기반으로 하는 각도이기 때문에 라디안 각도로 변환한다.
            double Dest_Lat_radian = P2_lat * (Math.PI / 180);
            double Dest_Lon_radian = P2_lon * (Math.PI / 180);

            // radian distance
            double radian_distance = 0;
            radian_distance = Math.Acos(Math.Sin(Cur_Lat_radian) * Math.Sin(Dest_Lat_radian) + Math.Cos(Cur_Lat_radian) * Math.Cos(Dest_Lat_radian) * Math.Cos(Cur_Lon_radian - Dest_Lon_radian));

            // 목적지 이동 방향을 구한다.(현재 좌표에서 다음 좌표로 이동하기 위해서는 방향을 설정해야 한다. 라디안값이다.
            double radian_bearing = Math.Acos((Math.Sin(Dest_Lat_radian) - Math.Sin(Cur_Lat_radian) * Math.Cos(radian_distance)) / (Math.Cos(Cur_Lat_radian) * Math.Sin(radian_distance)));        // acos의 인수로 주어지는 x는 360분법의 각도가 아닌 radian(호도)값이다.        

            double true_bearing = 0;
            if (Math.Sin(Dest_Lon_radian - Cur_Lon_radian) < 0)
            {
                true_bearing = radian_bearing * (180 / Math.PI);
                true_bearing = 360 - true_bearing;
            }
            else
            {
                true_bearing = radian_bearing * (180 / Math.PI);
            }

            // 반환값은 정북 기준으로 P1에서 P2를 바라보는 방향(deg)이다.
            return true_bearing;
        }

        public double LatLonDistance(double P1_lat, double P1_lon, double P2_lat, double P2_lon)
        {

            if ((P1_lat == P2_lat) && (P1_lon == P2_lon))
            {
                return 0;
            }
            double e10 = P1_lat * Math.PI / 180;
            double e11 = P1_lon * Math.PI / 180;
            double e12 = P2_lat * Math.PI / 180;
            double e13 = P2_lon * Math.PI / 180;
            /* 타원체 GRS80 */
            double c16 = 6356752.314140910;
            double c15 = 6378137.000000000;
            double c17 = 0.0033528107;
            double f15 = c17 + c17 * c17;
            double f16 = f15 / 2;
            double f17 = c17 * c17 / 2;
            double f18 = c17 * c17 / 8;
            double f19 = c17 * c17 / 16;
            double c18 = e13 - e11;
            double c20 = (1 - c17) * Math.Tan(e10);
            double c21 = Math.Atan(c20);
            double c22 = Math.Sin(c21);
            double c23 = Math.Cos(c21);
            double c24 = (1 - c17) * Math.Tan(e12);
            double c25 = Math.Atan(c24);
            double c26 = Math.Sin(c25);
            double c27 = Math.Cos(c25);
            double c29 = c18;
            double c31 = (c27 * Math.Sin(c29) * c27 * Math.Sin(c29))
                + (c23 * c26 - c22 * c27 * Math.Cos(c29))
                * (c23 * c26 - c22 * c27 * Math.Cos(c29));
            double c33 = (c22 * c26) + (c23 * c27 * Math.Cos(c29));
            double c35 = Math.Sqrt(c31) / c33;
            double c36 = Math.Atan(c35);
            double c38 = 0;
            if (c31 == 0)
            {
                c38 = 0;
            }
            else
            {
                c38 = c23 * c27 * Math.Sin(c29) / Math.Sqrt(c31);
            }
            double c40 = 0;
            if ((Math.Cos(Math.Asin(c38)) * Math.Cos(Math.Asin(c38))) == 0)
            {
                c40 = 0;
            }
            else
            {
                c40 = c33 - 2 * c22 * c26
                    / (Math.Cos(Math.Asin(c38)) * Math.Cos(Math.Asin(c38)));
            }
            double c41 = Math.Cos(Math.Asin(c38)) * Math.Cos(Math.Asin(c38))
                * (c15 * c15 - c16 * c16) / (c16 * c16);
            double c43 = 1 + c41 / 16384
                * (4096 + c41 * (-768 + c41 * (320 - 175 * c41)));
            double c45 = c41 / 1024 * (256 + c41 * (-128 + c41 * (74 - 47 * c41)));
            double c47 = c45
                * Math.Sqrt(c31)
                * (c40 + c45
                / 4
                * (c33 * (-1 + 2 * c40 * c40) - c45 / 6 * c40
                * (-3 + 4 * c31) * (-3 + 4 * c40 * c40)));
            double c50 = c17
                / 16
                * Math.Cos(Math.Asin(c38))
                * Math.Cos(Math.Asin(c38))
                * (4 + c17
                * (4 - 3 * Math.Cos(Math.Asin(c38))
                * Math.Cos(Math.Asin(c38))));
            double c52 = c18
                + (1 - c50)
                * c17
                * c38
                * (Math.Acos(c33) + c50 * Math.Sin(Math.Acos(c33))
                * (c40 + c50 * c33 * (-1 + 2 * c40 * c40)));
            double c54 = c16 * c43 * (Math.Atan(c35) - c47);

            // return distance in meter
            return c54;
        }

        public void ECEF2LatLonAlt(double x, double y, double z, ref double lat, ref double lon, ref double alt)
        {
            

	        double nla0,nla1,nla2,nla3,nla4,nlb0,nlb1,nlb2,nlb3;
	        double nlb,nlc0,nlopqk,nlqk,nlqkc,nlqks,nlf,nlfprm;
	        long nlk;
	        double ytemp, xtemp;
	        /*­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­*/
	        /* b = (x*x + y*y) / semi_major_axis_squared */
	        /* c = (z*z) / semi_major_axis_squared */
	        /*­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­*/
	        nlb = (x * x + y * y) / ASQR;
	        nlc0 = z * z / ASQR;
	        /*­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­*/
	        /* a0 = c * one_minus_eccentricity_sqr */
	        /* a1 = 2 * a0 */
	        /* a2 = a0 + b - first_eccentricity_to_fourth*/
	        /* a3 = ­2.0 * first_eccentricity_to_fourth */
	        /* a4 = - first_eccentricity_to_fourth */
	        /*­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­*/
	        nla0 = OMES * nlc0;
	        nla1 = 2.0 * nla0;
	        nla2 = nla0 + nlb - EFOR;
	        nla3 = -2.0 * EFOR;
	        nla4 = -EFOR;
	        /*­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­*/
	        /* b0 = 4 * a0, b1 = 3 * a1 */
	        /* b2 = 2 * a2, b3 = a3 */
	        /*­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­*/
	        nlb0 = 4.0 * nla0;
	        nlb1 = 3.0 * nla1;
	        nlb2 = 2.0 * nla2;
	        nlb3 = nla3;
	        /*­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­*/
	        /* Compute First Eccentricity Squared */
	        /*­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­*/
	        nlqk = ESQR;
	        for(nlk = 1; nlk <= 3; nlk++)
	        {
		        nlqks = nlqk * nlqk;
		        nlqkc = nlqk * nlqks;
		        nlf = nla0 * nlqks * nlqks + nla1 * nlqkc + nla2 * nlqks +
			        nla3 * nlqk + nla4;
		        nlfprm= nlb0 * nlqkc + nlb1 * nlqks + nlb2 * nlqk + nlb3;
		        nlqk = nlqk - (nlf / nlfprm);
	        }
	        /*­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­*/
	        /* Compute latitude, longitude, altitude */
	        /*­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­­*/
	        nlopqk = 1.0 + nlqk;
	        if ( x == 0.0 && y == 0.0 )/* on the earth's axis
											           */
	        {
		        /* We are sitting EXACTLY on the earth's axis */
		        /* Probably at the center or on or near one of the poles */
		        lon = 0.0; /* as good as any other value */
		        if(z >= 0.0)
			        lat = Math.PI / 2; /* alt above north pole */
		        else
                    lat = -(Math.PI) / 2; /* alt above south pole */
	        }
	        else
	        {
		        ytemp = nlopqk * z;
                xtemp = Math.Sqrt(x * x + y * y);
                lat = Math.Atan2(ytemp, xtemp);
                lon = Math.Atan2(y, x);
	        }

	        alt = (1.0 - OMES / ESQR * nlqk) * MAJA *
                Math.Sqrt(nlb / (nlopqk * nlopqk) + nlc0);

            lat *= 180.0 / Math.PI;
            lon *= 180.0 / Math.PI;
        }
        #endregion


        #region 시간 경과
        // 클럭타이머를 이용한 경과시간 체크 모듈
        public double GetDuration(DateTime start_time) // 입력 인자 start_time 으로부터 몇초가 지났는지를 반환.
        {
            double duration = 0.0;

            DateTime m_CurrentTime = new DateTime();
            m_CurrentTime = DateTime.Now;
            long elapsedTicks = m_CurrentTime.Ticks - start_time.Ticks; ;
            TimeSpan t_span = new TimeSpan(elapsedTicks);
            duration = (double)t_span.TotalSeconds;

            return duration;
        }
        #endregion


        #region 공유메모리
        // 파일 매핑 공유 메모리 생성
        public void CreateSharedMemory(ref string SharedMemKey)
        {
            try
            {
                mapFile = MemoryMappedFile.CreateNew(SharedMemKey, 2048);

                mapFile = MemoryMappedFile.OpenExisting(SharedMemKey, MemoryMappedFileRights.ReadWrite);
                accessor = mapFile.CreateViewAccessor();
            }
            catch (System.Exception ex){}
        }

        // 파일 매핑 공유 메모리 해제
        public void CloseSharedMem()
        {
            if (accessor != null)
            {
                accessor.Dispose();
                accessor = null;
            }

            if (mapFile != null)
            {
                mapFile.Dispose();
                mapFile = null;
            }
        }

        // 파일 매핑한 공유 메모리 쓰기
        public void WriteSharedMemory(ref string SharedMemKey, ref byte[] data)
        {
            try
            {
                // Mapping된 file 가져오기
                mapFile = MemoryMappedFile.OpenExisting(SharedMemKey, MemoryMappedFileRights.ReadWrite);
                accessor = mapFile.CreateViewAccessor();

                // 공유 Memory에 쓰기
                byte[] WriteData = new byte[2048];
                Buffer.BlockCopy(data, 0, WriteData, 0, data.Length);
                accessor.WriteArray(0, WriteData, 0, WriteData.Length);
            }
            catch (System.Exception ex)
            {
                Trace.WriteLine("생성된 공유메모리가 없음.");
            }            
        }

        // 파일 매핑한 공유 메모리 읽기
        public void ReadSharedMemory(ref string SharedMemKey, ref byte[] data)
        {
            string SharedMemString = null;
            try
            {
                MemoryMappedFile mapFile = MemoryMappedFile.OpenExisting(SharedMemKey, MemoryMappedFileRights.ReadWrite);

                using (Stream view = mapFile.CreateViewStream())
                {
                    // stream을 String으로 변환
                    view.Position = 0;
                    using (StreamReader reader = new StreamReader(view, Encoding.UTF8))
                    {
                        SharedMemString = reader.ReadToEnd();
                    }
                }
                data = Encoding.UTF8.GetBytes(SharedMemString); 
            }
            catch (System.Exception ex)
            {
                Trace.WriteLine("생성된 공유메모리가 없음.");
            }
            
        }

        
        #endregion


        #region 곡선 그리기
        public void getBezier2DPointList(ref List<PointF> points, int PointsOnCurve, ref List<PointF> result)
        {
            POINTS_ON_CURVE = PointsOnCurve * 2;

            double[] ptind = new double[points.Count * 2];
            double[] p = new double[POINTS_ON_CURVE];

            int idx = 0;
            for (int i = 0; i < points.Count * 2; i += 2)
            {
                ptind[i] = points[idx].X;
                ptind[i+1] = points[idx].Y;
                idx++;
            }

            bc.Bezier2D(ptind, (POINTS_ON_CURVE) / 2, p);

            for (int i = 0; i < POINTS_ON_CURVE; i += 2)
            {
                PointF currP = new PointF();
                currP.X = (float)p[i];
                currP.Y = (float)p[i + 1];

                result.Add(currP);
            }
        }

        public void getCatmullRom2DPointList(ref List<PointF> points, ref List<PointF> result)
        {
            var keyPoints = new List<PointF>();

            for (int i = 0; i < points.Count; i++)
            {
                keyPoints.Add(points[i]);
                if (i == 0)
                {
                    // 캣멀롬 곡선은 첫점과 마지막점이 시작과 끝곡선의 방향벡터로 들어가기 때문에
                    // 첫점과 끝점을 두번씩 입력해줘야 함.
                    keyPoints.Add(points[i]);
                }
            }
            keyPoints.Add(points[points.Count - 1]);

            var splinePoints = CatmullRomSpline.Generate(keyPoints.ToArray(), 10);

            for (int i = 0; i < splinePoints.Length; i++)
            {
                result.Add(splinePoints[i]);
            }
        }
        #endregion

        #region GetRealIpAddress
        public IPAddress GetRealIpAddress()
        {
            IPAddress gateway = FindGetGatewayAddress();

            if (gateway == null)
                return null;

            IPAddress[] pIPAddress = Dns.GetHostAddresses(Dns.GetHostName());

            foreach (IPAddress address in pIPAddress)
                if (IsAddressOfGateway(address, gateway))
                    return address;

            return null;
        }

        private bool IsAddressOfGateway(IPAddress address, IPAddress gateway)
        {
            if (address != null && gateway != null)
                return IsAddressOfGateway(address.GetAddressBytes(), gateway.GetAddressBytes());

            return false;
        }

        private bool IsAddressOfGateway(byte[] address, byte[] gateway)
        {
            if (address != null && gateway != null)
            {
                int gwLen = gateway.Length;
                if (gwLen > 0)
                {
                    if (address.Length == gateway.Length)
                    {
                        --gwLen;
                        int counter = 0;
                        for (int i = 0; i < gwLen; i++)
                        {
                            if (address[i] == gateway[i])
                                ++counter;
                        }
                        return (counter == gwLen);
                    }
                }
            }
            return false;
        }

        private IPAddress FindGetGatewayAddress()
        {
            IPGlobalProperties ipGlobProps = IPGlobalProperties.GetIPGlobalProperties();

            foreach (NetworkInterface ni in NetworkInterface.GetAllNetworkInterfaces())
            {
                IPInterfaceProperties ipInfProps = ni.GetIPProperties();
                foreach (GatewayIPAddressInformation gi in ipInfProps.GatewayAddresses)
                    return gi.Address;
            }
            return null;
        }
        #endregion

        
    }
}
